GLFW_ICON ICON "raylib.ico"

1 VERSIONINFO
FILEVERSION     5,5,0,0
PRODUCTVERSION  5,5,0,0
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    //BLOCK "080904E4"  // English UK
    BLOCK "040904E4"    // English US
    BEGIN
      VALUE "CompanyName", "raylib technologies"
      VALUE "FileDescription", "raylib dynamic library (www.raylib.com)"
      VALUE "FileVersion", "5.5.0"
      VALUE "InternalName", "raylib.dll"
      VALUE "LegalCopyright", "(c) 2024 Ramon Santamaria (@raysan5)"
      VALUE "OriginalFilename", "raylib.dll"
      VALUE "ProductName", "raylib"
      VALUE "ProductVersion", "5.5.0"
    END
  END
  BLOCK "VarFileInfo"
  BEGIN
    //VALUE "Translation", 0x809, 1252  // English UK
    VALUE "Translation", 0x409, 1252    // English US
  END
END
